--[[
Add, remove, lock, change target.
Update bars.

DToX, 2014
Licensed under the same terms as Lua itself.
--]]





local isDebug = true;
local isDebug = false;
function wlDebug( ... )
   if ( isDebug and wl ) then
      wl( ... );
   end
end

local debugBarValue = true;
local debugBarValue = false;
local function debugUpdateBarValue( barName, mark, vOld, vNew, c, ct, m, mt )
   local str = "Upd. " .. barName .. " value " .. mark .. ": " .. vOld .. " -> " .. vNew;
   str = str .. " (C: " .. c .. ", CT: " .. ct .. "; M: ".. m .. ", MT: " .. mt .. ")"
   if ( debugBarValue and wl ) then
      wl( str );
   end
end





-- Variables and constants
--
-- Defaults
local MAXIMUM_VALUE = 1500000000000; -- Some numbers are wrong and too big

local player = Turbine.Gameplay.LocalPlayer.GetInstance();
local classes = {}; -- Pairs ID = Name



-- Forward declarations
--





-- Functions
--

-- Returns new target's number
Vitals.GetNextNumber = function( self )
   local n = 1;
   for _ in ipairs( self.Object ) do
      n = n + 1;
   end
   return n;
end

-- Returns number of targets
Vitals.GetCount = function( self )
   local n = 0;
   for _ in pairs( self.Object ) do
      n = n + 1;
   end
   return n;
end

-- Removes specified target or all targets except specified
Vitals.RemoveTarget = function( self, i, mode )

wlDebug( "!!Vitals.RemoveTarget #", i );

   -- Remove all except specified
   if ( mode ) then
      for j in pairs( self.Object ) do
         if ( j ~= i ) then
            self.Object[j]:destroy();
         end
      end
   -- Remove specified
   else
      self.Object[i]:destroy();
   end

end

-- Locks target or gets new target. Returns boolean (true = locked)
Vitals.LockTarget = function( self, num )
   local objTarget = self.Object[num].Target;
   local isLocked = not objTarget.IsLocked;
   objTarget.IsLocked = isLocked;

wlDebug( "!!Vitals.LockTarget. #", num, ". IsLocked: ", isLocked );

   -- Save link
   if ( isLocked ) then
      objTarget.Locked = player:GetTarget();
   -- Update target
   else
      objTarget.Locked = nil;
      player:TargetChanged();
   end

   return isLocked;
end


-- Converts number to string and inserts specified delimiter, returns string
local function insertDelim( num, delim )
   local str = tostring( num );

   str, i = str:gsub( "(%d+)(%d%d%d)$", "%1" .. delim .. "%2" );
   while i > 0 do
      str, i = str:gsub( "(%d+)(%d%d%d)" .. delim, "%1" .. delim .. "%2" .. delim );
   end

   return str;
end

-- Returns function's value (rounded number) or 0, if function doesn't exist/some error
local function getFuncValue( obj, funcName )
   local ok, v = pcall( obj[funcName], obj );
   if ( ok ) then
      -- Fixing -1.#QNAN value. -1/-1.#QNAN (0%)
      local str = tostring( v );
      if ( str:find( "QNAN" ) ) then
         -- v = 0;
         v = str:gsub( ".#QNAN", "" ); -- -> -1
         -- v = str:gsub( "%D", "" ); -- -> 1
wlDebug( "!!getFuncValue. Fixing -1.#QNAN -> ", v );

      end

      -- Round to nearest
      return math.floor( v + 0.5 );
   else
      return 0;
   end
end

local function cNeg( v )
   if ( v < 0 ) then
      return 0;
   else
      return v;
   end
end

-- Gets current, maximum and temporary values, calculates % and sets bar's new value
local function updateBarValue( objTarget, num, barName )

   -- Exit 
   if ( not objTarget ) then
      return;
   end

   local curTarget = objTarget.Current;
   if ( not curTarget ) then
      return;
   end


   -- Get values
   local curValue = getFuncValue( curTarget, "Get" .. barName );
   local maxValue = getFuncValue( curTarget, "GetMax" .. barName );
   local curTempValue = getFuncValue( curTarget, "GetTemporary" .. barName );
   local maxTempValue = getFuncValue( curTarget, "GetMaxTemporary" .. barName );
   -- Correct negative values
   curValue = cNeg( curValue );
   maxValue = cNeg( maxValue );
   curTempValue = cNeg( curTempValue );
   maxTempValue = cNeg( maxTempValue );

   local current = curValue + curTempValue;
   local maximum = maxValue + maxTempValue;

   -- Correct current value
   if ( current > maximum and maximum < MAXIMUM_VALUE ) then
      local old = current; -- For debug
      current = maximum;
debugUpdateBarValue( barName, "current 2", old, current, curValue, curTempValue, maxValue, maxTempValue );
   end


   -- Correct maximum value
   if ( maximum > MAXIMUM_VALUE and current < MAXIMUM_VALUE ) then
      local old = maximum; -- For debug
      maximum = current;
debugUpdateBarValue( barName, "maximum 2", old, maximum, curValue, curTempValue, maxValue, maxTempValue );
   end

   -- Calculate percent, update text values and bar's background size
   local percent = 0;
   local backgroundWidth;
   if ( maxValue > 0 ) then
      if ( maximum ~= 0 ) then
         percent = current / maximum * 100;
      end

      backgroundWidth = math.floor( Vitals.Settings.Bar.Width * percent / 100 );
      percent = string.format( "%.1f", percent );
   end

   local objVitals = Vitals.Object[num];
   local currentStr = insertDelim( current, "." );
   local maximumStr = insertDelim( maximum, "." );
   local str = currentStr .. "/" .. maximumStr .. " (" .. percent .. "%)";
   objVitals:updateBar( barName, str, backgroundWidth );

   -- Unlock target. Sometimes works
   local isTargetLocked = objTarget.IsLocked;
   if ( isTargetLocked and barName == "Morale" and current <= 0 ) then

wlDebug( "Unlocking #", num, ". Current: ", current, ". Max: ", maximum, ". IsLocked: ", isTargetLocked );

      objVitals:toggleLock();
      player:TargetChanged();
   end

end

-- Updates values of all targets' bar.
-- To resolve and smooth some problems and target's updates
local function updateBarsValue( barName )

wlDebug( "!!updateBarsValue. Enter" );

   for i, objVitals in pairs( Vitals.Object ) do
      updateBarValue( objVitals.Target, i, barName );
   end
end

-- Used in other files.
-- Sets or resets target's values. Adds morale/power change functions
function targetChanged( num )

wlDebug( "!!targetChanged. Start" );

   local objVitals = Vitals.Object[num];
   local objTarget = objVitals.Target;
   local curTarget;

   -- Exit if target is locked
   if ( objTarget.IsLocked ) then

wlDebug( "!!targetChanged. #", objTarget.number, ". IsLocked: ", objTarget.IsLocked, ". Target is locked. Return." );

      return;
   end

wlDebug( "!!targetChanged. #", objTarget.number, ". Target is not locked. Continue." );

   -- Set current target and target's entity
   objTarget.Current = player:GetTarget();
   curTarget = objTarget.Current;
   objTarget.Entity:SetEntity( objTarget.Current );

   -- Clear bars
   objVitals:resetBar( "Morale" );
   objVitals:resetBar( "Power" );

   -- Player has something selected
   if ( curTarget ) then

      -- Get some general values
      local name = curTarget:GetName();
      local class = getFuncValue( curTarget, "GetClass" ); -- Returns ID
      class = classes[class];
      local level = getFuncValue( curTarget, "GetLevel" );

      local str = name;

      if ( level > 0 ) then
         str = "[" .. level .. "] " .. name;

         -- Get current values
         updateBarValue( objTarget, num, "Morale" );
         updateBarValue( objTarget, num, "Power" );

         -- Morale's value changes
         function curTarget:MoraleChanged()

wlDebug( "!!curTarget:MoraleChanged. #", num );

            updateBarsValue( "Morale" );
         end
         curTarget.MaxMoraleChanged = curTarget.MoraleChanged;
         curTarget.TemporaryMoraleChanged = curTarget.MoraleChanged;
         curTarget.MaxTemporaryMoraleChanged = curTarget.MoraleChanged;

         -- Power's value changes
         function curTarget:PowerChanged()

wlDebug( "!!curTarget:PowerChanged. #", num );

            updateBarsValue( "Power" );

-- wl( "IsLocal: ", self:IsLocalPlayer(), ". #", num );
         end
         curTarget.MaxPowerChanged = curTarget.PowerChanged;

      end

      -- Set name string and class icon
      objVitals:updateBar( "Name", str );
      objVitals:updateBar( "Class", class );

   -- No current target
   else

wlDebug( "!!end. # ", objTarget.number, ". No target. Clear all fields." );

      objVitals:updateBar( "Name", T( 05 ) );
      objVitals:updateBar( "Class" );
   end

end

-- Updates all targets
function player:TargetChanged()
   for i in pairs( Vitals.Object ) do
      targetChanged( i );
   end
end





-- Logic
--

-- Get currently available classes
for className, classID in pairs( Turbine.Gameplay.Class ) do
   if ( className ~= "IsA" ) then
      classes[classID] = className;
   end
end
