--[[
Create visual elements.

DToX, 2014
Licensed under the same terms as Lua itself.
--]]





-- Variables and constants
--

-- Backgrounds
local Images = {};
Images.Lock = { -- 20*20
   ["Open"] = 0x410001D3;
   ["Closed"] = 0x410001D1;
};
Images.Class = { -- Classes; 20*20
   ["Burglar"] = 0x41007DEB;
   ["Captain"] = 0x41007DE7;
   ["Champion"] = 0x41007DEC;
   ["Guardian"] = 0x41007DE8;
   ["Hunter"] = 0x41007DEA;
   ["LoreMaster"] = 0x41007DE9;
   ["Minstrel"] = 0x41007DE6;
   ["RuneKeeper"] = 0x410EA238;
   ["Warden"] = 0x410EA239;
   ["BlackArrow"] = 0x41007DEF;
   ["Defiler"] = 0x410E6BF4;
   ["Reaver"] = 0x41007DED;
   ["Stalker"] = 0x41007DF1;
   ["WarLeader"] = 0x41007DF0;
   ["Weaver"] = 0x41007DEE;
   ["Troll"] = 0x41091DE2;
   ["Ranger"] = 0x41091DE3;
   ["Chicken"] = 0x41091DE1;
};

-- Used in submenu
local fontList = {}; -- Pairs <i> = <font's name>
local languageList = {
   -- "English";
   -- "French";
   -- "German";
   -- "Russian";
   "ENG";
   "FRE";
   "GER";
   "RUS";
};


-- Default values
-- local BAR_WIDTH = 162; -- + padding * 4
-- local BAR_HEIGHT = 11; -- + padding * 2
local BAR_PADDING = 2; -- Between bar's text and border
local BAR_OFFSET = 1; -- Between bars
local BAR_MAXIMUM_TEXT = "1,999,999,999/1,999,999,999 (100.0%)";
local FONT_NAME = "Verdana12";
local FONT_NAME_BAR_NAME = "Verdana14";
local MENU_ITEM_OFFSET = "  ";
local TARGET_START_X = "650";
local TARGET_START_Y = "150";
local NEXT_TARGET_START_OFFSET = "5";

-- Vitals = {}; -- In Main.lua
-- Vitals.Settings = {}; -- In Main.lua
Vitals.Settings.Bar = {};
Vitals.Settings.Bar.Width = nil;
Vitals.Settings.Bar.Height = nil;
Vitals.Settings.Bar.Padding = BAR_PADDING;
Vitals.Settings.Bar.Offset = BAR_OFFSET;
Vitals.Settings.Bar.MaximumText = BAR_MAXIMUM_TEXT; -- For calculating bar's width
Vitals.Settings.Bar.Font = FONT_NAME; -- For morale/power bar
Vitals.Settings.Bar.FontNameBar = FONT_NAME_BAR_NAME; -- For name bar
Vitals.Object = {}; -- Pairs <number> = carrier

-- Context menu.
-- One menu to rule them all... And to prevent memory leak
local Menu = {};
Menu.Object = Turbine.UI.ContextMenu();
Menu.Items = Menu.Object:GetItems();
Menu.isCreated = false;
Menu.Number = nil; -- Target's number
Menu.ItemOffset = MENU_ITEM_OFFSET; -- Left border offset
Menu.Item = {}; -- For easier manipulation
Menu.Item.RemoveTarget = nil;
Menu.Item.RemoveTargets = nil;



-- Forward declarations
--





-- Functions
--

-- Creates and returns element with specified and some predifined properties
local function newLabel( parent, font )
   local label = Turbine.UI.Label();
   label:SetParent( parent );
   label:SetFont( font );
   label:SetTextAlignment( align[5] );
   return label;
end

local function newControl( parent, color )
   local control = Turbine.UI.Control();
   control:SetParent( parent );
   control:SetBackColor( color );
   return control;
end

-- Adds to one object ability to move another object on the screen
local function addDragging( obj, objToMove )
   local dragging = false;
   local dragStartX;
   local dragStartY;

   function obj:MouseDown( args )
      if ( args.Button == Turbine.UI.MouseButton.Left ) then
         dragStartX = args.X;
         dragStartY = args.Y;
         dragging = true;
      end
   end

   function obj:MouseMove( args )
      if ( dragging ) then
         local left, top = objToMove:GetPosition();
         local newLeft = left + ( args.X - dragStartX );
         local newTop = top + ( args.Y - dragStartY );
         objToMove:SetPosition( newLeft, newTop );
      end
   end

   function obj:MouseUp( args )
      if ( args.Button == Turbine.UI.MouseButton.Left ) then
         dragging = false;
      end
   end
end

local function getFont( varName )
   local fontName = Vitals.Settings.Bar[varName];
   return fonts[fontName] or fontName;
end

-- Adds new vitals window to the screen. Selects current target
Vitals.New = function( self )
   local num = self:GetNextNumber();
   local font = getFont( "Font" );
   local fontNameBar = getFont( "FontNameBar" );

   -- Contains all objects
   local carrier = Turbine.UI.Window();
   carrier:SetVisible( true );
   local o = num * NEXT_TARGET_START_OFFSET;
   carrier:SetPosition( TARGET_START_X + o, TARGET_START_Y + o );
   -- Hide visuals with all interface
   carrier:SetWantsKeyEvents( true );
   function carrier:KeyDown( args )
      local a = args.Action;
      if ( a == 268435635 ) then
         local visible = carrier:IsVisible();
         carrier:SetVisible( not visible );
      end
   end


   -- Entity for target (selection and RMB actions)
   local selectTarget = Turbine.UI.Lotro.EntityControl();
   selectTarget:SetParent( carrier );

   -- Name row
   local lbl_targetName = newLabel( carrier, fontNameBar );
   lbl_targetName:SetTextAlignment( align[4] );
   lbl_targetName:SetBackColor( colors.Black );
   lbl_targetName:SetMultiline( false );
   -- Name row. Show hint, if name's too long
   function lbl_targetName.MouseEnter( self )
      local labelHeight = self:GetHeight();
      local nameWidth = Vitals.Settings.Bar.Width - labelHeight;
      local text = self:GetText();
      local font = self:GetFont();
      local actualNameWidth = Text:getSize( text, font, "Single" );
      if ( actualNameWidth > nameWidth ) then
         Hint( self, text );
      end
   end
   -- Name row. Show context menu
   function lbl_targetName:MouseClick( args )
      local btnPressed = args.Button;
      local btnExpected = Turbine.UI.MouseButton.Right;
      if ( btnPressed == btnExpected ) then
         -- Disable ability to remove, if there is only one target
         local remTarget = Menu.Item.RemoveTarget;
         local remTargets = Menu.Item.RemoveTargets;
         local n = Vitals:GetCount();
         if ( n > 1 ) then
            remTarget:SetEnabled( true );
            remTargets:SetEnabled( true );
         else
            remTarget:SetEnabled( false );
            remTargets:SetEnabled( false );
         end

         -- Set target's number and show menu
         Menu.Number = num;
         Menu.Object:ShowMenu();
      end
   end

   -- Name row. Lock icon
   local lockBackground = newControl( carrier, colors.Black );
   local lock = newControl( carrier );
   lock:SetBackground( Images.Lock.Open );
   -- More accurate on resizes
   lock:SetMouseVisible( false );
   function lockBackground:MouseClick()
      carrier:toggleLock();
   end

   -- Morale row
   local targetMoraleBackground = newControl( carrier, colors.ForestGreen );
   -- Morale row. Text
   local lbl_targetMorale = newLabel( carrier, font );

   -- Power row
   local targetPowerBackground = newControl( carrier, colors.RoyalBlue );
   -- Power row. Text
   local lbl_targetPower = newLabel( carrier, font );

   -- Class icon
   local class = newControl( carrier );


   -- Add dragging
   targetMoraleBackground:SetMouseVisible( false );
   lbl_targetMorale:SetMouseVisible( false );
   targetPowerBackground:SetMouseVisible( false );
   lbl_targetPower:SetMouseVisible( false );
   class:SetMouseVisible( false );
   addDragging( selectTarget, carrier );
   addDragging( lbl_targetName, carrier );


   carrier.Number = num;
   -- Objects for easier manipulation
   carrier.Bar = {};
   carrier.Bar.Name = {};
   carrier.Bar.Name.Text = lbl_targetName;
   -- carrier.Bar.Name.Background = nil;
   carrier.Bar.Morale = {};
   carrier.Bar.Morale.Text = lbl_targetMorale;
   carrier.Bar.Morale.Background = targetMoraleBackground;
   carrier.Bar.Power = {};
   carrier.Bar.Power.Text = lbl_targetPower;
   carrier.Bar.Power.Background = targetPowerBackground;
   carrier.Icon = {};
   carrier.Icon.Lock = lock;
   carrier.Icon.Class = class;
   carrier.Target = {};
   carrier.Target.Current = nil;
   carrier.Target.Locked = nil;
   carrier.Target.IsLocked = false;
   carrier.Target.Entity = selectTarget;


   -- Clears specified bar's text and resets background's width
   carrier.resetBar = function( self, barName )
      local obj = self.Bar[barName];
      obj.Text:SetText( "" );
      obj.Background:SetWidth( Vitals.Settings.Bar.Width );
   end

   -- Updates specified bar's text and background or icon's background
   carrier.updateBar = function( self, barName, barText, barWidth )
      -- Class icon
      if ( barName == "Class" ) then
         local image = Images.Class[barText];
         local obj = self.Icon[barName];
         obj:SetBackground( image );
         -- Stretch image
         if ( image ) then
            obj:SetStretchMode( 2 );
         -- Prevent object's size change
         else
            obj:SetStretchMode( nil );
         end
      -- Bars
      else
         local obj = self.Bar[barName];
         obj.Text:SetText( barText );
         if ( barName ~= "Name" ) then
            local barWidth = barWidth or Vitals.Settings.Bar.Width;
            obj.Background:SetWidth( barWidth );
         end
      end
   end

   -- Locks or unlocks current target
   carrier.toggleLock = function( self )
      local num = self.Number;
      local objLock = self.Icon.Lock;
      local isLocked = Vitals:LockTarget( num );
      if ( isLocked ) then
         objLock:SetBackground( Images.Lock.Closed );
      else
         objLock:SetBackground( Images.Lock.Open );
      end
   end

   -- Changes size of all elements
   carrier.changeSize = function( self, newFont )
--[[
-- local lbl_targetName = self.Bar.Name.Text;
-- local lbl_targetMorale = self.Bar.Morale.Text;
-- local targetMoraleBackground = self.Bar.Morale.Background;
-- local lbl_targetPower = self.Bar.Power.Text;
-- local targetPowerBackground = self.Bar.Power.Background;
-- local selectTarget = self.Target.Entity;
--]]

      local W = Vitals.Settings.Bar.Width;
      local H = Vitals.Settings.Bar.Height;
      local P = Vitals.Settings.Bar.Padding;
      local O = Vitals.Settings.Bar.Offset;
      local nameH = H;
      local carrierH;

      -- Calculate new width and height, if it's first call or new font.
      -- Add padding, offsets, save new dimensions
      if ( not W or newFont ) then
         local font = getFont( "Font" );
         local maxText = Vitals.Settings.Bar.MaximumText;
         W, H = Text:getSize( maxText, font, "Single" );
         W = W + P * 5;
         H = H + P * 2;
         nameH = H;
         Vitals.Settings.Bar.Width = W;
         Vitals.Settings.Bar.Height = H;

         -- Check name bar
         local font = getFont( "FontNameBar" );
         local text = lbl_targetName:GetText();
         if ( text == "" ) then
            text = "T1";
         end
         local nW, nH = Text:getSize( text, font, "Single" );
         nameH = nH + P * 2;
      end

      carrierH = nameH + H * 2 + O * 2;

      carrier:SetSize( W, carrierH );
      selectTarget:SetSize( W, carrierH );

      lbl_targetName:SetSize( W - nameH, nameH );

      -- Lock
      local b = getBorders( lbl_targetName );
      lockBackground:SetPosition( b.Right, b.Top );
      lockBackground:SetSize( nameH, nameH );
      lock:SetSize( nameH, nameH );
      lock:SetStretchMode( 2 );
      lock:SetPosition( b.Right, b.Top );

      -- Morale
      targetMoraleBackground:SetSize( W, H );
      local b = getBorders( lbl_targetName );
      targetMoraleBackground:SetPosition( b.Left, b.Bottom + O );
      lbl_targetMorale:SetSize( W, H );
      lbl_targetMorale:SetPosition( b.Left, b.Bottom + O );

      -- Power
      targetPowerBackground:SetSize( W, H );
      local b = getBorders( lbl_targetMorale );
      targetPowerBackground:SetPosition( b.Left, b.Bottom + O );
      lbl_targetPower:SetSize( W, H );
      lbl_targetPower:SetPosition( b.Left, b.Bottom + O );

      -- Class icon
      local classOffset;
      if ( H > 20 ) then
         class:SetSize( H, H );
         classOffset = H / 2;
      else
         class:SetSize( 20, 20 );
         classOffset = 10;
      end
      local b = getBorders( targetMoraleBackground );
      local left = b.Left - classOffset;
      local top = b.Bottom - classOffset;
      class:SetPosition( left, top );
      class:SetStretchMode( 2 );
   end

   -- Changes specified bar's font
   carrier.changeBarFont = function( self, barName )
      local font;
      if ( barName == "Name" ) then
         font = getFont( "FontNameBar" );
      else
         font = getFont( "Font" );
      end

      local obj = self.Bar[barName].Text;
      local text = obj:GetText();
      obj:SetFont( font );
      obj:SetText( text );
   end

   -- Destroys carrier and vitals objects
   carrier.destroy = function( self )
      local num = self.Number;

      for k, v in pairs( self ) do
         carrier[k] = nil;
      end

      Vitals.Object[num] = nil;
   end

   -- Calculate size and check position
   carrier:changeSize( "New" );
   onScreen( carrier );

   Vitals.Object[num] = carrier; 

   -- Get current target
   targetChanged( num );

   -- Create context menu
   if ( not Menu.isCreated ) then
      Menu:create();
      Menu.isCreated = true;
   end

end

-- Changes bars' font and resizes all elements
Vitals.Redraw = function( self, nameBar )
   for _, obj in pairs( self.Object ) do
      obj:changeSize( "New" );

      if ( nameBar ) then
         obj:changeBarFont( "Name" );
      else
         obj:changeBarFont( "Morale" );
         obj:changeBarFont( "Power" );
      end
   end
end



-- Creates context menu
Menu.create = function( self )
   -- Context menu 01. Title. Disabled
   self:addItem( false );

   -- Context menu 02. Add target's vitals
   local menuItem = self:addItem();
   function menuItem:Click()
      Vitals();
   end

   -- Context menu 03. Remove target's vitals
   local menuItem = self:addItem();
   function menuItem.Click()
      local num = self.Number;
      Vitals:RemoveTarget( num );
   end
   self.Item.RemoveTarget = menuItem;

   -- Context menu 04. Remove all other target's vitals
   local menuItem = self:addItem();
   function menuItem.Click()
      local num = self.Number;
      Vitals:RemoveTarget( num, "All" );
   end
   self.Item.RemoveTargets = menuItem;

   -- Context menu 05. Settings. Title. Disabled
   self:addItem( false );

   -- Context menu. Settings. Font
   -- Context menu 06. Settings. Font (name bar). Submenu
   self:addFontSubmenu( "Name" );
   -- Context menu 07. Settings. Font (morale/power bar). Submenu
   self:addFontSubmenu();

   -- Context menu 08. Settings. Language
   self:addLanguageSubmenu();

   -- Context menu 09. Settings. Save settings
   local menuItem = self:addItem();
   function menuItem:Click()
      saveData();
   end

   self:updateText();
end

-- Adds new item to the context menu and returns it for additional tweaks
Menu.addItem = function( self, bEnabled )
   local item = Turbine.UI.MenuItem();
   if ( bEnabled ~= nil ) then
      item:SetEnabled( bEnabled );
   end

   self.Items:Add( item );

   return item;
end

-- Creates bar's font selection. Changes bars' font and size
Menu.addFontSubmenu = function( self, nameBar )
   local fontVar;
   if ( nameBar ) then
      fontVar = "FontNameBar";
   else
      fontVar = "Font";
   end

   local menuItem = self:addItem();
   local fontItems = menuItem:GetItems();
   fontItems.checked = nil;

   local subMenuItem;
   for i, fontName in pairs( fontList ) do
      subMenuItem = Turbine.UI.MenuItem( fontName ); -- +22 KB
      function subMenuItem.Click( self )
         -- Uncheck previous item and check current
         fontItems.checked:SetChecked( false );
         fontItems.checked = self;
         self:SetChecked( true );

         -- Save new font and update bars' font and size
         Vitals.Settings.Bar[fontVar] = fontName;
         Vitals:Redraw( nameBar );
      end
      fontItems:Add( subMenuItem );

      -- Check current item
      if ( Vitals.Settings.Bar[fontVar] == fontName ) then
         subMenuItem:SetChecked( true );
         fontItems.checked = subMenuItem;
      end
   end
end

-- Creates plugin's language selection. Changes translation
Menu.addLanguageSubmenu = function( self )
   local menuItem = self:addItem();
   local languageItems = menuItem:GetItems();
   languageItems.checked = nil;

   local subMenuItem;
   for i, langCode in pairs( languageList ) do
      subMenuItem = Turbine.UI.MenuItem( langCode );
      function subMenuItem.Click( self )
         -- Uncheck previous item and check current
         languageItems.checked:SetChecked( false );
         languageItems.checked = self;
         self:SetChecked( true );

         -- Change translation
         Vitals.Settings.Language = langCode;
         T( nil, true );
         Menu:updateText();
      end
      languageItems:Add( subMenuItem );

      -- Check current item
      if ( Vitals.Settings.Language == langCode ) then
         subMenuItem:SetChecked( true );
         languageItems.checked = subMenuItem;
      end
   end
end


-- Updates menu items' text and offset
Menu.updateText = function( self )
   -- <menu item's number> = { <text>, "offset multiplier" }
   local text = {
      [1] = { T( 10 ), 0 };
      [2] = { T( 15 ), 1 };
      [3] = { T( 20 ), 1 };
      [4] = { T( 25 ), 1 };
      [5] = { T( 30 ), 1 };
      [6] = { T( 35 ), 2 };
      [7] = { T( 40 ), 2 };
      [8] = { T( 45 ), 2 };
      [9] = { T( 50 ), 2 };
   };

   local items = self.Items;
   local n = items:GetCount();
   local offset = self.ItemOffset;
   local rep = string.rep;
   local item;
   for i = 1, n do
      item = items:Get( i );
      item:SetText( rep( offset, text[i][2] ) .. text[i][1] );
   end
end





-- Logic
--

-- To use Vitals instead of Vitals.New
local mt = {};
mt.__call = function( ... )
   return Vitals.New( ... );
end
setmetatable( Vitals, mt );

-- Get and sort currently available fonts
for font in pairs( Turbine.UI.Lotro.Font ) do
   if ( font ~= "IsA" and font ~= "Undefined" ) then
      local i = #fontList + 1;
      fontList[i] = font;
   end
end
table.sort( fontList );
